/* 

                          Firewall Builder

                 Copyright (C) 2002 NetCitadel, LLC

  Author:  Vadim Kurland     vadim@vk.crocodile.org

  $Id: NATCompiler_ipt.hh,v 1.22 2002/11/27 07:04:59 vkurland Exp $

  This program is free software which we release under the GNU General Public
  License. You may redistribute and/or modify this program under the terms
  of that license as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  To get a copy of the GNU General Public License, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef __NATCOMPILER_IPT_HH__
#define __NATCOMPILER_IPT_HH__

#include <fwbuilder/libfwbuilder-config.h>

#include "fwcompiler/NATCompiler.hh"
#include "fwbuilder/RuleElement.hh"

#include <map>
#include <string>

namespace libfwbuilder {
    class Host;
    class IPService;
    class ICMPService;
    class TCPService;
    class UDPService;
    class RuleElementOSrc;
    class RuleElementODst;
    class RuleElementOSrv;
    class RuleElementTSrc;
    class RuleElementTDst;
    class RuleElementTSrv;
};

namespace fwcompiler {

    class NATCompiler_ipt : public NATCompiler {

	protected:


        virtual std::string debugPrintRule(libfwbuilder::Rule *rule);

	/**
	 *  verifies correctness of the NAT rules
	 */
        DECLARE_NAT_RULE_PROCESSOR(VerifyRules);

	/**
	 *  verifies correctness of the NAT rules - this one
	 *  specifically * checks for some inconsistencies between
	 *  OSrv and TSrv and should * only be used after
	 *  splitServices
	 */
        DECLARE_NAT_RULE_PROCESSOR(VerifyRules2);

	/**
	 *  splits rule with multiple service objects in OSrv * onto
	 *  several rules
	 */
        DECLARE_NAT_RULE_PROCESSOR(splitOnOSrv);

	/**
	 *  process special case: rule that translates dest. port but
	 *  does not change addresses (tsrc==any, tdst==any, tsrv!=any)
	 *  Need to copy odst to tdst.
	 */
        DECLARE_NAT_RULE_PROCESSOR(portTranslationRules);

	/**
	 *  processor portTranslationRules copies ODst into TDst
         *  for rules that only do port translations and where TDst
	 *  is "any". In case ODst was firewall or one of its interfaces,
	 *  we should set rule type to NATRule::REDIRECT
	 */
        DECLARE_NAT_RULE_PROCESSOR(specialCaseWithRedirect);



        class splitRuleIfRuleElementIsDynamicInterface : public NATRuleProcessor
        {
            std::string re_type;
            public:
            splitRuleIfRuleElementIsDynamicInterface(const std::string &n,std::string _type):
                NATRuleProcessor(n) { re_type=_type; }
            virtual bool processNext();
        };

	/**
	 *  splits rule if one of the objects in tsrc is * interface
	 *  with dynamic address
	 */
        class splitOnDynamicInterfaceInTSrc : public splitRuleIfRuleElementIsDynamicInterface
        {
            public:
            splitOnDynamicInterfaceInTSrc(const std::string &n):
                splitRuleIfRuleElementIsDynamicInterface(n,libfwbuilder::RuleElementTSrc::TYPENAME) {}
        };

	/**
	 *  splits rule if one of the objects in odst is * interface
	 *  with dynamic address
	 */
        class splitOnDynamicInterfaceInODst : public splitRuleIfRuleElementIsDynamicInterface
        {
            public:
            splitOnDynamicInterfaceInODst(const std::string &n):
                splitRuleIfRuleElementIsDynamicInterface(n,libfwbuilder::RuleElementODst::TYPENAME) {}
        };

	/**
	 * checks for the following situations: 
         *
         *  1. dynamic interface is in OSrc and rule rtype is Masq or SNAT
         *  (drop interface from src since source address is
         *  undertermined)
         *
         *  2. dynamic interface is in ODst and rule type is DNAT
         *  (drop interface since dest. address is undefined)
         * 
	 */
        friend class specialCaseWithDynInterface;
        class specialCaseWithDynInterface : public NATRuleProcessor
        {
            void dropDynamicInterface(libfwbuilder::RuleElement *re);
            public:
            specialCaseWithDynInterface(const std::string &name) : NATRuleProcessor(name) {}
            virtual bool processNext();
        };


	/**
	 *  fills translated service with the copy of original srv
	 */
        DECLARE_NAT_RULE_PROCESSOR(fillTranslatedSrv);

	/**
	 *  assigns NAT rules to interfaces 
	 */
        DECLARE_NAT_RULE_PROCESSOR(AssignInterface);

	/**
	 *  calls OSConfigurator to add virtual * address to the
	 *  firewall if it is needed for NAT rule
	 */
        DECLARE_NAT_RULE_PROCESSOR(addVirtualAddress);

	/**
	 *   replaces references to the firewall in odst * with
	 *   references to its external interfaces
	 */
        DECLARE_NAT_RULE_PROCESSOR(ReplaceFirewallObjectsODst);

	/**
	 *   replaces references to the firewall in tsrc * with
	 *   references to its external interfaces
	 */
        DECLARE_NAT_RULE_PROCESSOR(ReplaceFirewallObjectsTSrc);

	/**
	 *  distinguishes SNAT from Masquerading (can do * this after
	 *  firewall objects has been replaced with its * interfaces
	 *  and basic NAT rule type has been determined)
	 */
        DECLARE_NAT_RULE_PROCESSOR(dynamicInterfaceInTSrc);

	/**
	 *  takes care of dynamic interfaces in ODst  (if ODst contains
	 *  interface and its address is dynamic, replace it with any)
	 */
        DECLARE_NAT_RULE_PROCESSOR(dynamicInterfaceInODst);

        /**
         *  splits rule element if src or dst contains * address
         *  range. This inspector differs from the standard one * in
         *  the base class NATCompiler
         */
        DECLARE_NAT_RULE_PROCESSOR(addressRanges);

	/**
	 * split rules with more than one service object, so that each
	 * rule has services with the same protocol
	 */
        DECLARE_NAT_RULE_PROCESSOR(splitServices);

	/**
	 * separate TCP/UDP services with port ranges (can not be used with multiport)
	 */
        DECLARE_NAT_RULE_PROCESSOR(separatePortRanges);

	/**
	 * for TCP/UDP services separate objects with source ports
         * from objects with destination ports
	 */
        DECLARE_NAT_RULE_PROCESSOR(separateSourcePorts);

	/**
	 * for TCP/UDP services separate objects that have both source
	 * and destination ports
	 */
        DECLARE_NAT_RULE_PROCESSOR(separateSourceAndDestinationPorts);

	/**
	 * splits rules so multiport module can be used
	 */
        DECLARE_NAT_RULE_PROCESSOR(prepareForMultiport);

        /**
         *  deals with negation in OSrc 
         */
        DECLARE_NAT_RULE_PROCESSOR(doOSrcNegation);

        /**
         *  deals with negation in ODst
         */
        DECLARE_NAT_RULE_PROCESSOR(doODstNegation);

        /**
         *  deals with negation in OSrv 
         */
        DECLARE_NAT_RULE_PROCESSOR(doOSrvNegation);

	/**
	 * splits NONAT rule and assigns chains to PREROUTING,
	 * POSTROUTING and OUTPUT. Always call this processor before
	 * decideOnChain
	 */
        DECLARE_NAT_RULE_PROCESSOR(splitNONATRule);


	/**
	 * decides what chain this rule should go to. 
	 */
        DECLARE_NAT_RULE_PROCESSOR(decideOnChain);

	/**
	 * decides on "jump to" chain
	 */
        DECLARE_NAT_RULE_PROCESSOR(decideOnTarget);

        /**
         *  eliminates duplicate objects in SRC. Uses default comparison
         *  in eliminateDuplicatesInRE which compares IDs
         */
        class eliminateDuplicatesInOSRC : public eliminateDuplicatesInRE
        {
            public:
            eliminateDuplicatesInOSRC(const std::string &n) :
                eliminateDuplicatesInRE(n,libfwbuilder::RuleElementOSrc::TYPENAME) {}
        };

        /**
         *  eliminates duplicate objects in DST. Uses default comparison
         *  in eliminateDuplicatesInRE which compares IDs
         */
        class eliminateDuplicatesInODST : public eliminateDuplicatesInRE
        {
            public:
            eliminateDuplicatesInODST(const std::string &n) :
                eliminateDuplicatesInRE(n,libfwbuilder::RuleElementODst::TYPENAME) {}
        };

        /**
         *  eliminates duplicate objects in SRV. Uses default comparison
         *  in eliminateDuplicatesInRE which compares IDs
         */
        class eliminateDuplicatesInOSRV : public eliminateDuplicatesInRE
        {
            public:
            eliminateDuplicatesInOSRV(const std::string &n) :
                eliminateDuplicatesInRE(n,libfwbuilder::RuleElementOSrv::TYPENAME) {}
        };


	/**
	 *   prints single policy rule, assuming all * groups have
	 *   been expanded, so source, destination and * service hold
	 *   exactly one object each, and this object is * not a
	 *   group.  Negation should also have been taken care of *
	 *   before this method is called.
	 */
        class PrintRule : public NATRuleProcessor
        {
            bool                             init;
            std::string                      current_rule_label;
            std::map<const std::string,bool> chains;

            virtual void _printProtocol(libfwbuilder::Service *srv);
            virtual void _printSrcService(libfwbuilder::RuleElementOSrv  *o);
            virtual void _printDstService(libfwbuilder::RuleElementOSrv  *o);

            virtual std::string _printICMP(libfwbuilder::ICMPService *srv);
            virtual std::string _printIP(libfwbuilder::IPService *srv);

            virtual std::string _printOPorts(int rs,int re);
            virtual std::string _printTPorts(int rs,int re);
            virtual std::string _printSrcPorts(libfwbuilder::Service *srv);
            virtual std::string _printDstPorts(libfwbuilder::Service *srv);
            virtual std::string _printSNATPorts(libfwbuilder::Service *srv);
            virtual std::string _printDNATPorts(libfwbuilder::Service *srv);

            virtual void _printMultiport(libfwbuilder::NATRule *r);

            virtual void _printAddr(libfwbuilder::Address  *o,bool print_mask=true,bool print_range=false);

            virtual void _printChainDirectionAndInterface(libfwbuilder::NATRule *r);

            public:
            PrintRule(const std::string &name);
            virtual bool processNext();
        };

	virtual std::string myPlatformName();

	public:

	NATCompiler_ipt(libfwbuilder::FWObjectDatabase *_db,
		       const std::string &fwname,
		       fwcompiler::OSConfigurator *_oscnf) : NATCompiler(_db,fwname,_oscnf) { }


	virtual int  prolog();
	virtual void compile();
	virtual void epilog();
	
	static std::string getNewTmpChainName(libfwbuilder::NATRule *rule);

    };


}

#endif
