/* 

                          Firewall Builder

                 Copyright (C) 2002 NetCitadel, LLC

  Author:  Vadim Kurland     vadim@vk.crocodile.org

  $Id: pf.cc,v 1.35 2002/10/20 23:42:58 vkurland Exp $

  This program is free software which we release under the GNU General Public
  License. You may redistribute and/or modify this program under the terms
  of that license as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  To get a copy of the GNU General Public License, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "config.h"

#ifdef HAVE_LOCALE_H
#include <locale.h>
#endif

#include <fstream>
#include <iostream>
#include <algorithm>
#include <functional>

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>

#include <assert.h>

#include "PolicyCompiler_pf.hh"
#include "NATCompiler_pf.hh"
#include "fwbuilder/Resources.hh"

#include "OSConfigurator_openbsd.hh"
#include "OSConfigurator_freebsd.hh"
#include "OSConfigurator_solaris.hh"

#include "fwbuilder/FWObjectDatabase.hh"
#include "fwbuilder/XMLTools.hh"
#include "fwbuilder/FWException.hh"
#include "fwbuilder/Firewall.hh"

#include "fwcompiler/manifest.hh"

#ifdef HAVE_GETOPT_H
  #include <getopt.h>
#else
  #ifdef __MINGW32__
    #include <getopt.h>
  #else
    #include <stdlib.h>
  #endif
#endif


static const char      *filename       = NULL;
static const char      *wdir           = NULL;
static const char      *mfile          = NULL;
static const char      *fwobjectname   = NULL;
static int              dl             = 0;
static int              drp            = -1;
static int              drn            = -1;
static int              verbose        = 0;

using namespace libfwbuilder;
using namespace fwcompiler;

class UpgradePredicate: public XMLTools::UpgradePredicate
{
    public:
    virtual bool operator()(const string &msg) const 
    { 
	cout << _("Data file has been created in the old version of Firewall Builder. Use fwbuilder GUI to convert it.") << endl;
	return false;
    }
};
    
void usage(const char *name)
{
    cout << _("Firewall Builder:  policy compiler for OpenBSD PF") << endl;
    cout << _("Version ") << VERSION << RELEASE_NUM << endl;
    cout << _("Usage: ") << name << " [-x] [-v] [-V] [-f filename.xml] [-d destdir] [-m manifest] firewall_object_name" << endl;
}

int main(int argc, char * const *argv)
{   

#ifdef ENABLE_NLS
    setlocale (LC_ALL, "");

    bindtextdomain (PACKAGE, LOCALEDIR);
    textdomain (PACKAGE);
#else
#  ifdef HAVE_SETLOCALE
    setlocale (LC_ALL, "");
#  endif
#endif
    

    if (argc<=1)
    {
        usage(argv[0]);
        exit(1);
    }

    int   opt;

    while( (opt=getopt(argc,argv,"x:zvVf:d:m:")) != EOF )
    {
        switch(opt)
        {
        case 'd':
            wdir = strdup(optarg);
            break;
        case 'm':
            mfile = strdup(optarg);
            break;
        case 'f':
            filename = strdup(optarg);
            break;
        case 'x':
            if (*optarg=='p') {
                ++optarg;
                drp  = atoi(optarg);
            } else {
                if (*optarg=='n') {
                    ++optarg;
                    drn  = atoi(optarg);
                } else {
                    if (isdigit(*optarg))  dl=atoi(optarg);  // increase debug level
                    else {
                        usage(argv[0]);
                        exit(1);
                    }
                }
            }
            break;
        case 'v':
            verbose++;
            break;
        case 'V':
            usage(argv[0]);
            exit(1);
        }
    }
    
    if((argc-1) != optind)
    {
        usage(argv[0]);
        exit(1);
    }

    fwobjectname = strdup( argv[optind++] );


    if (wdir==0) 	wdir="./";

    if(chdir(wdir))   {
	cerr << _("Can't change to: ") << wdir << endl;
	exit(1);
    }


    try   {

        libfwbuilder::init();

        new Resources(TEMPLATE_DIR);

	/* create database */
	new FWObjectDatabase();

	/* load the data file */
	UpgradePredicate upgrade_predicate; 

	if (verbose) cout << _(" *** Loading data ...");
	FWObjectDatabase::db->load(filename,  &upgrade_predicate);
	if (verbose) cout << _(" done\n");


	string   fw_file_name=string(fwobjectname)+".fw";
	ofstream fw_file(fw_file_name.c_str());

	/* Review firewall and OS options and generate commands */
	Firewall*  fw=FWObjectDatabase::db->findFirewallByName(fwobjectname);
	FWOptions* options=fw->getOptionsObject();
	string s;

	string firewall_dir=options->getStr("firewall_dir");
	if (firewall_dir=="") firewall_dir="/etc";

	bool debug=options->getBool("debug");
	string shell_dbg=(debug)?"-x":"" ;
	string pfctl_dbg=(debug)?"-v":"";

	fw_file << "#!/bin/sh " << shell_dbg << endl;

/*
 * Process firewall options, build OS network configuration script
 */
	OSConfigurator *oscnf=NULL;
	if (fw->getStr("host_OS")=="solaris")   oscnf=new OSConfigurator_solaris(FWObjectDatabase::db , fwobjectname);
	if (fw->getStr("host_OS")=="openbsd")   oscnf=new OSConfigurator_openbsd(FWObjectDatabase::db , fwobjectname);
	if (fw->getStr("host_OS")=="freebsd")   oscnf=new OSConfigurator_freebsd(FWObjectDatabase::db , fwobjectname);
	if (oscnf==NULL)
	    throw FWException("Unrecognized host OS "+fw->getStr("host_OS"));

	oscnf->prolog();


/*
 * create compilers and run the whole thing 
 */
        Manifest manifest;
	Action a;


	PolicyCompiler_pf c( FWObjectDatabase::db , fwobjectname , oscnf );

	c.setDebugLevel( dl );
	c.setDebugRule(  drp );
	c.setVerbose( verbose );

	string   pf_file_name=string(fwobjectname)+"-pf.conf";
	ofstream pf_file(pf_file_name.c_str());
	bool     have_pf=false;
	if ( c.prolog() > 0 ) {

	    have_pf=true;

	    c.compile();
	    c.epilog();

	    a.action     = "COPY";
	    a.filename   = pf_file_name;
	    a.parameters = firewall_dir;
	    manifest.push_back(a);
	}


	NATCompiler_pf n( FWObjectDatabase::db , fwobjectname , oscnf );

	n.setDebugLevel( dl );
	n.setDebugRule(  drn );
	n.setVerbose( verbose );

	string   nat_file_name=string(fwobjectname)+"-nat.conf";
	ofstream nat_file(nat_file_name.c_str());
	bool     have_nat;
	if ( n.prolog() > 0 ) {

	    have_nat=true;

	    n.compile();
	    n.epilog();

	    a.action     = "COPY";
	    a.filename   = nat_file_name;
	    a.parameters = firewall_dir;
	    manifest.push_back(a);
	}
/*
 * now write generated scripts to files
 */
	fw_file << oscnf->getCompiledScript();

	if (have_pf) {
	    pf_file << c.getCompiledScript();
	    pf_file.close();
	    fw_file << "$PFCTL " << pfctl_dbg << " -R " << pf_file_name << endl;
	}

	if (have_nat) {
	    nat_file << n.getCompiledScript();
	    nat_file.close();
	    fw_file << "$PFCTL " << pfctl_dbg << " -N " << nat_file_name <<  endl;
	}

	fw_file << endl;

	fw_file << "$PFCTL -e" << endl;



	if(mfile)
            manifest.save(string(mfile), S_IRUSR|S_IWUSR);
        
        chmod(fw_file_name.c_str(),S_IXUSR|S_IRUSR|S_IWUSR|S_IRGRP);
        
        return 0;

    } catch(FWException &ex)  {
	cerr << ex.toString() << endl;
        return 1;
    } catch (std::string s) {
	cerr << s;
        return 1;
    } catch (std::exception ex) {
	cerr << ex.what();
        return 1;
    } catch (...) {
	cerr << _("Unsupported exception");
        return 1;
    }

}









